@@ -1,8 +1,11 @@
 #ifndef _SYNC_HOOKS_H_
 #define _SYNC_HOOKS_H_
 
-struct nlnetwork;
-struct us_conntrack;
+#include <sys/select.h>
+
+struct nethdr;
+struct cache_object;
+struct fds;
 
 struct sync_mode {
 	int internal_cache_flags;
@@ -13,13 +16,13 @@ struct sync_mode {
 	int  (*init)(void);
 	void (*kill)(void);
 	int  (*local)(int fd, int type, void *data);
-	int  (*pre_recv)(const struct nlnetwork *net);
-	void (*post_send)(int type,
-			  const struct nlnetwork *net,
-			  struct us_conntrack *u);
+	int  (*recv)(const struct nethdr *net);
+	void (*enqueue)(struct cache_object *obj, int type);
+	void (*xmit)(void);
 };
 
-extern struct sync_mode notrack;
-extern struct sync_mode nack;
+extern struct sync_mode sync_alarm;
+extern struct sync_mode sync_ftfw;
+extern struct sync_mode sync_notrack;
 
 #endif
